--[[
Copyright (C) 2016  Neemys

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

See <LICENSE.txt> for a copy of this license, or see <http://www.gnu.org/licenses/>
]]


require("misc")
require("config")

--will store platform tile to layer relation to check in collsion_mask if a building has been restricted with by a specific tile. As platforms do not block each other buildings using collision mask, it will help checking if a building is blocked by a platform using script check
platformToLayer = nil


--Using the dummy entities that we created in data-final-fixes phase, it will populate the platformToLayer variable for the script to use.
local function fillPlatformToLayer(game)
	platformToLayer = {}
	for _,tileProto in pairs(game.tile_prototypes) do
		if(game.entity_prototypes[buildingplatform.constants.dummyPrefix..tileProto.name]) then
			platformToLayer[tileProto.name] = game.entity_prototypes[buildingplatform.constants.dummyPrefix..tileProto.name].order
		end
	end
end



-- Exposed in the interface "building-platform"
-- Given an existing entity, check if it is blocked by a platform or by the terrain using rules set in the game loading.
-- Return false if the entity is wrongly placed, true otherwise.
-- Warning : This function do not actually remove the entity. If we return false and do nothing, under certain circonstance - like if concrete under the building is removed - the game's collision_mask check migth see a collision and make the building disappear without firing an destruction event or error message. 
checkGoodBuildingPlacement = function(entity)
	if not entity then return false end
	if not platformToLayer then fillPlatformToLayer(game) end --as we don't have access to game in control nor on_load, we have no choice but generate it here if needed
	local box = entity.bounding_box
	local isGhost = (entity.name == "entity-ghost");
	local entityPrototype = (isGhost and entity.ghost_prototype or entity.prototype );
	local collisionMask = entityPrototype.collision_mask
	local isRestrictedBuilding = false
	--check if it is a restricted buildings on any platform
	for _,layer in pairs(platformToLayer) do
		if(collisionMask[layer]) then
			isRestrictedBuilding = true
		end
	end
	if isRestrictedBuilding then--obviously we only need to check restricted building
		for x=box.left_top.x,box.right_bottom.x,1 do
			for y=box.left_top.y,box.right_bottom.y,1 do
				local tilePrototype = entity.surface.get_tile(x,y).prototype
				local hiddenTileName = entity.surface.get_hidden_tile({x,y}) --get_hidden_tile return only the name, not the tile
				local hiddenTilePrototype 
				if(hiddenTileName) then
					hiddenTilePrototype = game.tile_prototypes[hiddenTileName]
				end
				--check for every layer in our building collsion mask
				for layer,active in pairs(collisionMask) do
					if(active) then
						if(tilePrototype.collision_mask[layer] or hiddenTilePrototype and hiddenTilePrototype.collision_mask[layer]) then
							return false
						end
					end
				end
				
				
				--check if the building shoudn't be placed on that particular platform as we don't block each other platform
				local visiblePlatformLayer =  platformToLayer[tilePrototype.name]
				--we only perform the ckeck if the actual tile or hidden tile is a platform and we perform the test on the tile and the hidden tile if there is a platform on top of each other both rules apply. We can't ovveride the hidden platform rule by the one on top, because in case a platform is added afterward by the player, the building will not be placeale again if removed
				if(visiblePlatformLayer and not collisionMask[visiblePlatformLayer]) then 
					return false
				else
					local hiddenPlatformLayer = hiddenTilePrototype and platformToLayer[hiddenTilePrototype.name] or nil
					if(hiddenPlatformLayer and not collisionMask[hiddenPlatformLayer]) then
						return false
					end
				end
			end
		end
	end
	return true
end

-- Exposed in the interface "building-platform"
-- Destroy the given entity and give back the building to the given player. The building is looted in case it already contain items.
-- If you want to show a custom error message to the player, you should set the third parameter to false.
-- Return true if everything went fine, false otherwise.
removeAndReimburseBuilding = function(entity,player,needMessage)
	if not entity then return false end
	if not player then return false end
	if not needMessage then needMessage = true end
	local isGhost = (entity.name == "entity-ghost");
	local entityPrototype = (isGhost and entity.ghost_prototype or entity.prototype );
	-- when removing, we need to loot it in case another mod put item inside
	-- need a solution if some item can't go inside inventory, maybe spawn a chest ? For the moment they just vanish
	if(not isGhost) then
		local loot = entityPrototype.loot
		if loot then
			for i,loot_ele in pairs(loot) do
				player.insert({name=loot.item,count=loot.count_max})
			end
		end
		for i,item in pairs(entity.prototype.mineable_properties.products) do
			player.insert({name=item.name})  -- The entity by itself is not included in loot list
		end
	end
	entity.destroy()
	--Tell the player why the building has been removed if asked to do it.
	if(needMessage) then
		player.print("This building cannot be build here !")
	end
	return true
end


-- Exposed in the interface "building-platform"
-- Given an existing entity, check if it is blocked by a platform or by the terrain using rules set in the game loading. If the entity is wrongly placed, it is removed and reimbursed to the given player with an error message in chat.
-- Return true if the building is correctly placed, false otherwise.
checkBuildingPlacementAndReimburseWithMessage = function(entity,player)
	local isGoodPlacement = checkGoodBuildingPlacement(entity)
	if(not isGoodPlacement) then
		removeAndReimburseBuilding(entity,player,true)
	end	
	return isGoodPlacement
end

--Expose our functions in the interface.
remote.add_interface("building-platform",{checkGoodBuildingPlacement=checkGoodBuildingPlacement,removeAndReimburseBuilding=removeAndReimburseBuilding,checkBuildingPlacementAndReimburseWithMessage=checkBuildingPlacementAndReimburseWithMessage})



-- When player build something, check if it is a restricted building and build on a platform. Remove it if it must be.
script.on_event(defines.events.on_built_entity, function(event)	
	checkBuildingPlacementAndReimburseWithMessage(event.created_entity,game.players[event.player_index])
end)


-- In case a recipe is activated midgame and the research is already done, seek where the recipe is unlocked, check if the research is done and unlock the recipe. This function is ressource heavy, shouldn't be called often. 
-- Can be used to hide the recipe if the recipe have been deactivated in config.
script.on_event("bp-reload-recipe", function(event)
	for _,force in pairs(game.forces) do
		if force.recipes[buildingplatform.constants.tileName] then -- only if the recipe exist
			if(buildingplatform.config.craftable) then 
				local technologyFound = false
				local recipeToUnlocked = false
				local function inEffect(recipeName,effectsArray)
					for _,effect in pairs(effectsArray) do
						if effect['type'] == "unlock-recipe" and effect['recipe'] == recipeName then
							return true
						end
					end
					return false
				end
				--check for building platform default config
				if(force.technologies["concrete"] and inEffect(buildingplatform.constants.tileName,force.technologies["concrete"].effects)) then
					technologyFound = true
					if force.technologies["concrete"].researched then
						recipeToUnlocked = true
					end
				end
				--if not in default config, search where it has been put
				if not technologyFound then
					for _,tech in pairs(force.technologies) do
						if inEffect(buildingplatform.constants.tileName,tech.effects) then
							technologyFound = true
							if tech.researched then
								recipeToUnlocked = true
								break --stop the loop only if the recipe is to be unlocked, if not, maybe another tech that unlock the platform is researched
							end
						end
					end
				end
				if recipeToUnlocked then
					force.recipes[buildingplatform.constants.tileName].enabled = true
				end
			else --no craftable 
				force.recipes[buildingplatform.constants.tileName].enabled = false --in case someone want to deactivate the recipe²²
			end
		end
	end
end)


-- m2cengine's concentric-circle platform generator code
if(buildingplatform.config.platformGenerator == "concentric-circle" and not buildingplatform.config.disableOwnPlatform) then
	local append = table.insert

	local function limit( a, b, v )
		return math.min( b, math.max( a, v ) )
	end

	local abs = math.abs
	local min = math.min
	local max = math.max
	local pow = math.pow
	local rand = math.random --factorio's math.random is deterministic so no desync in multi (https://forums.factorio.com/viewtopic.php?p=46431#p46431)
	local floor = math.floor

	local function concentricCircleGenerate( x, y )
		ret = false
		ret2 = false
		
		r = math.sqrt( x * x + y * y )
		if math.floor( r ) % 150 < limit( 3, 16, 3 + .15 * r * .3 ) and rand() < .85 then
			ret = true
		end
		rmd = r * .01 * 1.0
		if rmd < 1 and rand() < 1.1 - pow( rmd, 4 ) then
			ret = true
		end
		if abs( x ) < 2 + 8 - .005 * abs( y ) and rand() < pow( 1 - .003 * r, .05 ) then ret = true end
		if abs( y ) < 2 + 8 - .005 * abs( x ) and rand() < pow( 1 - .003 * r, .05 ) then ret = true end
		
		if abs( x ) < 3 or abs( y ) < 3 then
			ret2 = true
		end
		
		if abs( x - y ) < 3 + .0 * r or abs( y + x ) < 3 + .0 * r then
			ret2 = true
		end
		
		if ( max( abs( x ), abs( y ) ) + 50 ) % 130 < 3 then
			ret2 = true
		end
		
		
		if rmd < .2 then
			ret = false
			ret2 = false
		end
		return ret, ret2
		
	end

	local function concentricCircleGenerateChunk(event)
		
		if event.surface.name ~= "nauvis" then --we only generate the platform for default surface (nauvis)
			return
		end

		local x1 = event.area.left_top.x
		local y1 = event.area.left_top.y
		local x2 = event.area.right_bottom.x
		local y2 = event.area.right_bottom.y
		local surface = event.surface
		local k = 5

		tiles = {}
		
		for x = x1 - k, x2 + k do
			for y = y1 - k, y2 + k do				
				local old = surface.get_tile(x, y).name
				new = nil
				g1, g2 = concentricCircleGenerate( x, y )
				
				if g1 == true then
					new = buildingplatform.constants.tileName
				end
				if g2 == true then
					new = buildingplatform.constants.speedTileName
					
				end
				if not (old == "water" or old == "deepwater") then --we do not place our tile on water
					if not (new == old) and new then
						append(tiles, {name = new, position = {x, y}})
					end	
				end
			end
		end
		surface.set_tiles(tiles)
	end

	script.on_event(defines.events.on_chunk_generated, concentricCircleGenerateChunk)
end