--[[
Copyright (C) 2016  Neemys

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

See <LICENSE.txt> for a copy of this license, or see <http://www.gnu.org/licenses/>
]]

require('internals_functions')

--[[ 
	Function that should be used by other mod if they want to integrate their own rules
	
	- Multiple pattern can be applied to a building type (call the function multiple time with different pattern and same type), a logical OR will be applied to all pattern.
	- If there is "block all but" and "block only" pattern for a building type, "block only" rules will be ignored.
	- Patterns are checked by using lua string.match function against the building name and should match the whole name, so format it the right way.
	- Patterns and building type should be strings.
	- No verification on building type is done, if you pass wrong building type, game will likely throw an error.
	- Building platform own config file does not filter other mod rules. If a user unblock boiler in config and you block it by blockAll("boiler"), it will be blocked, you have access to what the user want to block or not in buildingplatform.config table (see config.lua for more information).
	- Those functions should be called between data-updates.lua phase and data-final-fixes of this mod. Before they do not exist, after they have no effect.
	- The tile object must be an object returned by addPlatform function. If not specified, it will be applied to all platform that requested to be in the default list.
	- If a building type is added to any platform rules, all restricted building outside of this platform will not be placeable on ground nor on other platform unless you add rules to another platform to accept them. I intend to add more flexibility on platforms rules in the future, but for the moment it's how it work.
]]


-- Block all building of a given type.
buildingplatform.blockAll = function(buildingType,tileObject)
	buildingType = buildingType or "";
	tileObject = tileObject or nil;
	buildingplatform.checkParameters(buildingType,"")
	local tileObjectTable = buildingplatform.getTileObjectTable(tileObject)
	for _,tObject in pairs(tileObjectTable) do
		if(not tObject.buildingsAllowed[buildingType]) then --only block all if no rule of blocking already exist
			tObject.buildingsAllowed[buildingType] = {}; --empty table mean all building blocked
		end
	end
end

--block all building a given type but no those which name match the pattern.
buildingplatform.blockAllBut = function(buildingType, pattern,tileObject)
	buildingType = buildingType or "";
	pattern = pattern or "";
	tileObject = tileObject or nil;
	buildingplatform.checkParameters(buildingType,pattern)
	local tileObjectTable = buildingplatform.getTileObjectTable(tileObject)
	for _,tObject in pairs(tileObjectTable) do
		if(not tObject.buildingsAllowed[buildingType]) then
			tObject.buildingsAllowed[buildingType] = {pattern,nil};
		else
			table.insert(tObject.buildingsAllowed[buildingType],pattern)
		end
	end
end

--block only building which name match the pattern.
buildingplatform.blockOnly = function(buildingType, pattern,tileObject)
	buildingType = buildingType or "";
	pattern = pattern or "";
	tileObject = tileObject or nil;
	buildingplatform.checkParameters(buildingType,pattern)
	local tileObjectTable = buildingplatform.getTileObjectTable(tileObject)
	for _,tObject in pairs(tileObjectTable) do
		if(not tObject.buildingsDisallowed[buildingType]) then
			tObject.buildingsDisallowed[buildingType] = {pattern}
		else
			table.insert(tObject.buildingsDisallowed[buildingType],pattern)
		end
	end
end

-- Don't want building platform rules to mess with yours ? Call this and building platform own rules won't be applied. This are for vanilla rules and mod compatibility rules. It will not only stop applying  blocking to our own platform but also to all platform created by other mod and added to the default initialization list.
-- Other mod custom rules will still apply. Remember that people sometime want some building to be blocked and some other don't. Try to integrate the config file directives.
buildingplatform.disableBuildingPlatformOwnBlockList = function()
	buildingplatform.config.disableOwnBlocking = true;
end


--add alias with underscore instead of camelCase for those who prefer them.
buildingplatform.block_all = buildingplatform.blockAll
buildingplatform.block_all_but = buildingplatform.blockAllBut
buildingplatform.block_only = buildingplatform.blockOnly
buildingplatform.disable_building_platform_own_block_list = buildingplatform.disableBuildingPlatformOwnBlockList